/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include "recovery.h"
#include "udpsocket.h"

#define PROT_UDP			17

#define MTYPE_CALL			0

#define RPC_VERS			2

#define PMAP_PROG			100000
#define PMAP_VERS			2
#define PMAP_PROC_GETPORT		3
#define PMAP_PROC_GETPORT_SEND_SIZE	56


#define MOUNT_PROG			100005
#define MOUNT_VERS			1
#define MOUNT_PROC_DUMP			2
#define MOUNT_PROC_DUMP_SEND_SIZE	40

#define U32_LENGTH			4

#define RECV_TIME_OUT_SEC		1
#define RECV_TIME_OUT_USEC		0

#define SEND_BUFF_SIZE			PMAP_PROC_GETPORT_SEND_SIZE
#define SEND_BUFF_SIZE_U32		((SEND_BUFF_SIZE + 3) >> 2)
#define RECV_BUFF_SIZE			(1024 << 4)
#define RECV_BUFF_SIZE_U32		((RECV_BUFF_SIZE + 3) >> 2)

#define RECV_REPLY_STAT			2
#define RECV_ACCEPT_STAT		5
#define RECV_RESULT_OFFSET		6


static unsigned int xid;

int
mdump_init( struct udp_socket *sock, unsigned long int addr )
{

	int	st, port = 1024;

	sock->addr.sin_family = PF_INET;
//	sock->addr.sin_addr.s_addr = htonl( INADDR_ANY );
	sock->addr.sin_addr.s_addr = addr;
	do
	{
		sock->addr.sin_port = htons(--port);
		st = udp_connect( sock );
	}
	while ( st == EINVAL && port );

	return st;
}

int
mdump_getport( struct udp_socket * sock )
{
	int	st;
	unsigned int *	work = (unsigned int *)sock->send.buf;

	*work++ = htonl( xid++ );
	*work++ = htonl( MTYPE_CALL );
	*work++ = htonl( RPC_VERS );
	*work++ = htonl( PMAP_PROG );
	*work++ = htonl( PMAP_VERS );
	*work++ = htonl( PMAP_PROC_GETPORT );
	*work++ = htonl( 0 );	/* cred auth null */
	*work++ = htonl( 0 );
	*work++ = htonl( 0 );	/* verf auth null */
	*work++ = htonl( 0 );
	*work++ = htonl( MOUNT_PROG );
	*work++ = htonl( MOUNT_VERS );
	*work++ = htonl( PROT_UDP );
	*work   = htonl( 0 );
	sock->send.len = PMAP_PROC_GETPORT_SEND_SIZE;
	if ( !( st = udp_send( sock ) ) )
	{
		struct timeval t;
#if 0
printf( "PMAP SEND OK\n" );
#endif
		t.tv_sec  = RECV_TIME_OUT_SEC;
		t.tv_usec = RECV_TIME_OUT_USEC;
		sock->recv.len = RECV_BUFF_SIZE;
		if ( ( st = udp_canreceive( sock, &t ) ) &&
			!( st = udp_receive( sock ) ) )
		{
#if 0
printf( "PMAP RECV OK\n" );
#endif
			work = (unsigned int *)sock->recv.buf;
			if ( sock->recv.len == 28 &&
				!( st = work[RECV_REPLY_STAT] ) &&
				!( st = work[RECV_ACCEPT_STAT] ) )
				sock->send.addr.sin_port =
					htons( (unsigned short)ntohl(work[6]) );
		}
	}
	return st;
}

int
mdump( struct udp_socket * sock )
{
	int	st;

	if ( !( st = mdump_getport( sock ) ) )
	{
		unsigned int *	work = (unsigned int *)sock->send.buf;
#if 0
printf( "GET PORT (%d)\n", ntohs(sock->send.addr.sin_port) );
#endif
		work[0] = htonl( xid++ );
		work[3] = htonl( MOUNT_PROG );
		work[4] = htonl( MOUNT_VERS );
		work[5] = htonl( MOUNT_PROC_DUMP );
		sock->send.len = MOUNT_PROC_DUMP_SEND_SIZE;
		if ( !( st = udp_send( sock ) ) )
		{
			struct timeval t;
#if 0
printf( "MOUNT SEND OK\n" );
#endif
			t.tv_sec  = RECV_TIME_OUT_SEC;
			t.tv_usec = RECV_TIME_OUT_USEC;
			sock->recv.len = RECV_BUFF_SIZE;
			if ( ( st = udp_canreceive( sock, &t ) ) )
				st = udp_receive( sock );
#if 0
if ( !st )
{
	work = (unsigned int *)sock->recv.buf;
	if ( !work[2] && !work[5] )
	{
		int	i = 6;
		printf( "PMAP RECV OK(%d)\n", sock->recv.len );
		sock->recv.len >>= 2;
		while ( i < sock->recv.len )
			printf( "%08X\n", ntohl( work[i++] ) );
	}
}
#endif
		}
	}
	return st;
}

static inline int
checkmount_one( char *host, unsigned int *buff, int len, char *path )
{
#if 0
printf( "[ %s : %s ] : %0X", host, path, ntohl(*buff) );
#endif

	if ( strlen( host ) == len && !memcmp( host, buff, len ) )
	{
		int off = ((len + 3) >> 2);
		len = htonl( buff[ off++ ] );
#if 0
printf( " : %0X\n", ntohl(buff[off]) );
#endif
		return ( strlen( path ) != len ||
			 memcmp( path, &buff[off], len ) );
	}
#if 0
printf( "\n" );
#endif

	return -1;
}

static inline int
checkmount( char **hosts, unsigned int *buff, char *path )
{
	int	st = -1, off = 0, len;
	char **	names;

	while ( st && buff[ off++ ] )
	{
		len = htonl( buff[ off++ ] );
		names = hosts;
		while( *names &&
			( st = checkmount_one(
					*names, &buff[off], len, path ) ) )
			*names++;
		off += ((len + 3) >> 2);
		len = htonl( buff[ off++ ] );
		off += ((len + 3) >> 2);
	}
#if 0
printf( "check mount(%d)\n", st );
#endif
	return st;
}

static inline char **
lookup_client( pool_t * pool, unsigned long addr )
{
	struct hostent *host;
	char **		name = NULL;

	if ( (host=gethostbyaddr( (const char *)&(addr),
				   sizeof(addr), AF_INET ) ) )
	{
		int i = 0, n = 1;
		char *	str;
		char **	work = host->h_aliases;

		while( *work++ ) n++;
		work = host->h_addr_list;
		while( *work++ ) n++;

		name = (char **)pool_alloc(
				pool, sizeof( char * ) * ( n + 1 ) );
		*( name + n ) = NULL;
		if ( !( *( name + i ) = pool_alloc(
				pool, strlen( host->h_name ) + 1 ) ) )
			goto err_out;
		strcpy( *( name + i++ ), host->h_name );
		work = host->h_aliases;
		while ( *work )
		{
			if ( !( *( name + i ) = pool_alloc(
				pool, strlen( *work ) + 1 ) ) )
				goto err_out;
			strcpy( *( name + i++ ), *work++ );
		}
		work = host->h_addr_list;
		while ( *work )
		{
			str = (char *)inet_ntoa( *(struct in_addr *)*work++ );
			if ( str && !( *( name + i ) =
				pool_alloc( pool, strlen( str ) + 1 ) ) )
				goto err_out;
			strcpy( *( name + i++ ), str );
		}
	}
	return name;
err_out:
	return NULL;
}

int
auth_client( char *export, unsigned long int sip, unsigned long int cip )
{
	int		st;
#if 0
	unsigned short	port = CLIENT_AUTH_PORT;
#else
	unsigned short	port = htons( pmap_port );
#endif
	unsigned int	send[SEND_BUFF_SIZE_U32];
	unsigned int	recv[RECV_BUFF_SIZE_U32];
	struct udp_socket
			sock;
	char **		clientnames;
	pool_t *	pool = create_pool( 4096 );

	if ( !pool ) return -1;
	clientnames = lookup_client( pool, cip );

	if ( clientnames && !( st = mdump_init( &sock, sip ) ) )
	{
#if 0
printf( "udp init ok\n" );
#endif
		sock.send.buf = (unsigned char *)send;
		sock.recv.buf = (unsigned char *)recv;
		sock.send.addr.sin_family = PF_INET;
		sock.send.addr.sin_addr.s_addr = sip;
		sock.send.addr.sin_port = port;

		st = mdump( &sock );
#if 0
printf( "dump(%d)\n", st );
#endif
		udp_disconnect( &sock );

		if ( !st && !( st = recv[RECV_REPLY_STAT] ) &&
			    !( st = recv[RECV_ACCEPT_STAT] ) )
			st = checkmount( clientnames,
				&recv[RECV_RESULT_OFFSET], export );
	}

	delete_pool( pool );

	return st;
}

