/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

static inline int
check_client( check_t * chk )
{
	return auth_client( chk->serv_ex, chk->serv_ip, chk->cip );
}

static inline int
command_auth( int fd, check_t * chk )
{
	chk->recvbuff.offset = 0;
	if ( chk->pathbuff.offset == 0 )
	{
		int		slen;
		struct stat	st;

		chk->recvbuff.length = to_bytelen( 3 );
		if ( receive_buffer( fd, &chk->recvbuff ) <
						to_bytelen( 3 ) )
			goto out_err;

		chk->cip     = get_nlong( &chk->recvbuff );
		chk->serv_ip = get_nlong( &chk->recvbuff );
		chk->serv_ex = chk->pathbuff.buffer;
		slen = get_hlong( &chk->recvbuff );

		chk->pathbuff.length = quad_u32len( slen );
		if ( receive_buffer( fd, &chk->pathbuff ) <
						quad_u32len( slen ) )
			goto out_err;

		/* UMZ 2002/05/10 */
		if (slen > 0 && chk->pathbuff.buffer[ slen - 1 ] == '/') {
		  chk->pathbuff.buffer[ slen - 1] = '\0';
		  chk->pathbuff.offset = chk->pathbuff.length = slen - 1;
		} else {
		  chk->pathbuff.buffer[ slen ] = '\0';
		  chk->pathbuff.offset = chk->pathbuff.length = slen;
		}

#if 0
fprintf( stdFp, "CLIENT : %s MOUNT : %s : ",
		inet_ntoa( chk->cip ), chk->serv_ex );
#endif
#if 1
		if ( stat( chk->serv_ex, &st ) == -1 ||
				check_client( chk ) )
			goto out_err;
#else
		if ( stat( chk->serv_ex, &st ) == -1 ||
				check_client( chk ) == -1 )
			goto out_err;
#endif
#if 0
fprintf( stdFp, "AUTH OK\n" );
#endif
		return send_status( fd,  STATUS_OK );
	}
out_err:
#if 0
fprintf( stdFp, "AUTH NG\n" );
#endif
	return -1;
}

static inline int
command_stat( int fd, check_t * chk )
{
	chk->targ_ct = 0;
	chk->recvbuff.offset = 0;
	if ( chk->serv_ex && chk->pathbuff.length > 0 )
	{
		int		slen;
		unsigned long	csum;

		chk->recvbuff.length = to_bytelen( 4 );
		if ( receive_buffer( fd, &chk->recvbuff ) <
						to_bytelen( 4 ) )
			goto out_err;
		chk->targ_ch = get_hlong( &chk->recvbuff );
		chk->targ_of = get_hlong( &chk->recvbuff );
		chk->targ_ln = get_hlong( &chk->recvbuff );

		slen = get_hlong( &chk->recvbuff );
		chk->recvbuff.offset = 0;
		chk->recvbuff.length = quad_u32len( slen );
		if ( receive_buffer( fd, &chk->recvbuff ) <
						quad_u32len( slen ) )
			goto out_err;

		chk->pathbuff.length = chk->pathbuff.offset;
		chk->targ_pt =
			&( chk->pathbuff.buffer[ chk->pathbuff.offset ] );
		*chk->targ_pt++ = '/';
		memcpy( chk->targ_pt, chk->recvbuff.buffer, slen );
		chk->pathbuff.length += ( slen + 1 );
		chk->pathbuff.buffer[ chk->pathbuff.length ] = '\0';

#if 1
fprintf( stdFp, "TARGET : %s : DATACHECK %s : OFFSET %d : LENGTH %d\n",
		chk->targ_pt,
		( chk->targ_ch ? "true" : "false" ),
		chk->targ_of, chk->targ_ln );
#endif
		csum = 0;
		readstat( chk->serv_ex, &chk->targ_type, &chk->targ_st );
		if ( chk->targ_type == TYPE_REG &&
			chk->targ_of == (~0) &&
			chk->targ_of == chk->targ_ln )
			chk->targ_of = 0;
		if ( chk->targ_ch )
			csum = checksum( chk->targ_type, chk->serv_ex,
						chk->targ_of, chk->targ_ln );

#if 0
fprintf( stdFp, "TARGET : %s : DATACHECK %s : OFFSET %d : LENGTH %d ",
		chk->targ_pt,
		( chk->targ_ch ? "true" : "false" ),
		chk->targ_of, chk->targ_ln );
fprintf( stdFp, "TYPE %d : MODE %X : UID %d : GID %d : SIZE %u : CSUM %X\n",
		chk->targ_type, chk->targ_st.st_mode,
		chk->targ_st.st_uid, chk->targ_st.st_gid,
		chk->targ_st.st_size, csum );
#endif

		chk->sendbuff.offset = chk->sendbuff.length = 0;
		if ( !free_u32length( &chk->sendbuff, 6 ) )
			goto out_err;
		set_hlong( &chk->sendbuff, chk->targ_type );
		set_hlong( &chk->sendbuff, chk->targ_st.st_mode );
		set_hlong( &chk->sendbuff, chk->targ_st.st_uid );
		set_hlong( &chk->sendbuff, chk->targ_st.st_gid );
		set_hlong( &chk->sendbuff, chk->targ_st.st_size );
		set_nlong( &chk->sendbuff, csum );
		return send_buffer( fd, &chk->sendbuff );
	}
out_err:
	return -1;
}

static inline unsigned long
get_stat( int fd, check_t * chk,
		char * path, unsigned long * t, struct stat * s )
{
	unsigned long bl;

	if ( fd == -1 || !path || !t || !s )
		goto out_err;

	chk->sendbuff.offset = chk->sendbuff.length = 0;
	set_hlong( &chk->sendbuff, COMMAND_STAT );
	set_hstr( &chk->sendbuff, path );
	bl = chk->sendbuff.length;
	if ( send_buffer( fd, &chk->sendbuff ) < bl )
		goto out_err;
	
	chk->recvbuff.offset = 0;
	chk->recvbuff.length = to_bytelen( 5 );
	if ( receive_buffer( fd, &chk->recvbuff ) < to_bytelen( 5 ) )
		goto out_err;

	*t = get_hlong( &chk->recvbuff );
	s->st_mode = get_hlong( &chk->recvbuff );
	s->st_uid  = get_hlong( &chk->recvbuff );
	s->st_gid  = get_hlong( &chk->recvbuff );
	s->st_size = get_hlong( &chk->recvbuff );

	return 0;
out_err:
	return -1;
}

static inline unsigned long
check_exists( int fd, check_t * chk,
		char * path, unsigned long * t, struct stat * s )
{
	unsigned long bl;

	if ( fd == -1 || !path || !t || !s )
		goto out_err;

	chk->sendbuff.offset = chk->sendbuff.length = 0;
	set_hlong( &chk->sendbuff, COMMAND_EXISTS );
	set_hstr( &chk->sendbuff, path );
	bl = chk->sendbuff.length;
	if ( send_buffer( fd, &chk->sendbuff ) < bl )
		goto out_err;
	
	chk->recvbuff.offset = 0;
	chk->recvbuff.length = to_bytelen( 5 );
	if ( receive_buffer( fd, &chk->recvbuff ) < to_bytelen( 5 ) )
		goto out_err;

	*t = get_hlong( &chk->recvbuff );
	s->st_mode = get_hlong( &chk->recvbuff );
	s->st_uid  = get_hlong( &chk->recvbuff );
	s->st_gid  = get_hlong( &chk->recvbuff );
	s->st_size = get_hlong( &chk->recvbuff );

	return 0;
out_err:
	return -1;
}

static inline unsigned long
next_entry( int fd, check_t * chk, unsigned long * o,
		unsigned long * t, struct stat * s, char * n )
{
	unsigned long bl, sl;

	if ( fd == -1 || !o || !t || !s || !n )
		goto out_err;

	chk->sendbuff.offset = chk->sendbuff.length = 0;
	set_hlong( &chk->sendbuff, COMMAND_ENTRY );
	set_hlong( &chk->sendbuff, *o );
	bl = chk->sendbuff.length;
	if ( send_buffer( fd, &chk->sendbuff ) < bl )
		goto out_err;
	
	chk->recvbuff.offset = 0;
	chk->recvbuff.length = to_bytelen( 7 );
	if ( receive_buffer( fd, &chk->recvbuff ) < to_bytelen( 7 ) )
		goto out_err;

	*t = get_hlong( &chk->recvbuff );
	s->st_mode = get_hlong( &chk->recvbuff );
	s->st_uid  = get_hlong( &chk->recvbuff );
	s->st_gid  = get_hlong( &chk->recvbuff );
	s->st_size = get_hlong( &chk->recvbuff );
	*o = get_hlong( &chk->recvbuff );
	sl = get_hlong( &chk->recvbuff );

	if ( tcp_read( fd, n, quad_u32len( sl ) ) < quad_u32len( sl ) )
		goto out_err;
	n[ sl ] = '\0';
#if 0
fprintf( stdFp, "[%d:%s][%d][%d][%d][%d][%d]\n",
		*o, n, *t, s->st_mode, s->st_uid, s->st_gid, s->st_size );
fprintf( stdFp, "CHECK TARGET ENTRY : %s OFFSET %u\n", n, *o );
#endif

	return 0;
out_err:
	return -1;
}

static inline int
repair_remove( char * path, unsigned long * t, struct stat * s )
{
#if 1
	fprintf( stdFp, "Remove : target %s ", path );
#endif

	if ( remove( path ) == -1 )
		goto out_err;
	*t = NO_EXISTS;
	memset( s, 0, sizeof( struct stat ) );
#if 1
	fprintf( stdFp, "done.\n" );
#endif
	return 0;
out_err:
#if 1
	fprintf( stdFp, "error.\n" );
#endif
	return -1;
}

static inline int
repair_attribute( char * path,
			unsigned long * t1, struct stat * s1,
				unsigned long t2, struct stat * s2 )
{
	int	fr = 0;

	switch( t2 )
	{
		case TYPE_REG:
			if ( s1->st_size != s2->st_size )
				if ( truncate( path,  s2->st_size ) == -1 )
				{
					fr = -1;
					goto out;
				}
				else
				{
					fr = 1;
				}
		break;

		case TYPE_LNK:
			s1->st_mode = s2->st_mode;
			s1->st_size = s2->st_size;
		break;

		case TYPE_DIR:
			s1->st_size = 0;
		break;

		default:
			fr -1;
			goto out;
		break;
	}

	if ( s1->st_mode != s2->st_mode )
	{
		if ( chmod( path, s2->st_mode ) != -1 )
		{
			fr = 1;
			s1->st_mode = s2->st_mode;
		}
		else
		{
			fr = -1;
			goto out;
		}
	}
	if ( s1->st_uid != s2->st_uid ||
		s1->st_gid != s2->st_gid )
	{
		if ( lchown( path, s2->st_uid, s2->st_gid ) != -1 )
		{
			fr = 1;
			s1->st_uid = s2->st_uid;
			s1->st_gid = s2->st_gid;
		}
		else
		{
			fr = -1;
			goto out;
		}
	}
	*t1 = t2;
out:
	if ( fr )
	{
		fprintf( stdFp, "Repair attribute : target %s %s\n",
				path, ( ( fr == 1 ) ? "done." : "error." ) );

	}
	return ( ( fr == 1 ) ? 0 : fr );
}

static inline int
repair_reg( char *path, int fd, check_t * chk, mode_t mode,
		unsigned long of, unsigned long ln )
{
	int		state = -1, wfd;
	unsigned long	rlen, len;

#if 1
	fprintf( stdFp, "Repair file : target %s ", path );
#endif
	chk->sendbuff.offset = chk->sendbuff.length = 0;
	set_hlong( &chk->sendbuff, COMMAND_DATA );
	set_hlong( &chk->sendbuff, of );
	set_hlong( &chk->sendbuff, ln );
	if ( send_buffer( fd, &chk->sendbuff ) < to_bytelen( 3 ) )
		goto out;

	if ( receive_status( fd, &rlen ) < 0 || rlen ==  STATUS_NG )
		goto out;

	if ( receive_status( fd, &rlen ) < 0 )
		goto out;
	
	if ( ( wfd = open( path, O_RDWR | O_CREAT, mode ) ) == -1 )
		goto out;
	if ( lseek( wfd, of, SEEK_SET ) == -1 )
		goto out_close;

	len = ( ( rlen < BUFFER_SIZE_BYTE ) ? rlen : BUFFER_SIZE_BYTE );

	while( len && ( len = tcp_read( fd, chk->recvbuff.buffer, len ) ) > 0 ) 
	{
		if ( write( wfd, chk->recvbuff.buffer, len ) < len )
			goto out_close;
		if ( ( rlen -= len ) <= 0 ) 
			break;
		len = ( ( rlen < len ) ? rlen : len );
	}
	state = 0;

out_close:
	close( wfd );
out:
#if 1
	fprintf( stdFp, "%s.\n", ( state ? "error" : "done" ) );
#endif
	return state;
}

static inline int
repair_link( char *path, int fd, check_t * chk,
		unsigned long of, unsigned long ln )
{
	int		state = -1;
	unsigned long	rlen;

#if 1
	fprintf( stdFp, "Repair link : target %s ", path );
#endif
	chk->sendbuff.offset = chk->sendbuff.length = 0;
	set_hlong( &chk->sendbuff, COMMAND_DATA );
	set_hlong( &chk->sendbuff, of );
	set_hlong( &chk->sendbuff, ln );
	if ( send_buffer( fd, &chk->sendbuff ) < to_bytelen( 3 ) )
		goto out;

	if ( receive_status( fd, &rlen ) < 0 || rlen ==  STATUS_NG )
		goto out;

	if ( receive_status( fd, &rlen ) < 0 )
		goto out;

	if ( rlen > ( BUFFER_SIZE_BYTE - 1 ) )
		goto out;

	chk->recvbuff.offset = 0;
	chk->recvbuff.length = rlen;

	if ( receive_buffer( fd, &chk->recvbuff ) < rlen )
		goto out;
	chk->recvbuff.buffer[ rlen ] = '\0';

	state = symlink( chk->recvbuff.buffer, path );
#if 1
	fprintf( stdFp, "done.\n" );
#endif
out:
#if 1
	fprintf( stdFp, "error.\n" );
#endif
	return state;
}

static inline int
path_down( check_t * chk, char * path )
{
        int     	tn = strlen( chk->serv_ex ), pn = strlen( path );
	unsigned long	st, bl;

        if ( ( tn + pn + 2 ) > BUFFER_SIZE_BYTE )
                return -1;
        chk->serv_ex[ tn++ ] = '/';
        strcpy( &( chk->serv_ex[ tn ] ), path );
        chk->targ_ct ++;

        return 0;
}

static inline int
path_up( check_t * chk )
{
        int     	tn = strlen( chk->serv_ex );
	unsigned long	st;

        if ( ( --( chk->targ_ct ) ) < 0 )        return -1;
        while ( tn > 0 && chk->serv_ex[ --tn ] != '/' );
        if ( tn < 0 )           return -1;
        chk->serv_ex[ tn ] = '\0';

        return 0;
}

static inline int
repair_create( char * path, unsigned long * t1, struct stat * s1,
		int fd, check_t * chk,
		unsigned long t2, struct stat * s2 );

static inline int
repair_data( char * path, unsigned long * t1, struct stat * s1,
		int fd,check_t * chk,
		unsigned long t2, struct stat * s2,
		unsigned long of, unsigned long ln );

static inline int
repair_entry( char * path, unsigned long * t1, struct stat * s1,
		int fd, check_t * chk,
		unsigned long t2, struct stat * s2 )
{
	if ( *t1 != t2 )
	{
		if ( *t1 != NO_EXISTS )
			repair_remove( path, t1, s1 );
		repair_create( path, t1, s1, fd, chk, t2, s2 );
	}
	else
	{
		if ( s1->st_size != s2->st_size || t2 == TYPE_DIR )
			repair_data( path, t1, s1, fd, chk, t2, s2, 0, (~0) );
		else
			repair_attribute( path, t1, s1, t2, s2 );
	}
	return 0;
}

int
check_entry( char *path, int fd, check_t * chk )
{

        DIR *           d;
        struct dirent * e;
        struct stat     st;
        unsigned long   ty;

        if ( !( d = opendir( path ) ) )
                goto out_err;

	while( ( e = readdir( d ) ) )
	{
		if ( !check_exists( fd, chk, e->d_name, &ty, &st ) &&
			ty == NO_EXISTS )
		{
			path_down( chk, e->d_name );
			fprintf( stdFp, "Remove : target %s ",
					chk->serv_ex );
		        if ( remove( chk->serv_ex ) != -1 )
				fprintf( stdFp, "done.\n" );
			else
				fprintf( stdFp, "error.\n" );
			path_up( chk );
		}
	}
	closedir( d );
	return 0;
out_err:
	return -1;
}

int
repair_dir( char *path, int fd, check_t * chk )
{
	int		state = -1;
	unsigned long	o = 0;
	unsigned long	t, nt;
	struct stat	s, ns;
	static char	n[256];

	while ( next_entry( fd, chk, &o, &t, &s, n ) != -1 && strlen( n ) )
	{
#if 0
fprintf( stdFp, "<<<%s[%u]>%X>%X:%d:%d:%u>\n",
		n, o, t, s.st_mode, s.st_uid, s.st_gid, s.st_size );
#endif
		path_down( chk, n );
#if 0
fprintf( stdFp, "DOWN >> PATH %s\n", path );
#endif
		if ( strcmp( n, "." ) && strcmp( n, ".." ) )
		{
			readstat( path, &nt, &ns );
			repair_entry( path, &nt, &ns, fd, chk, t, &s );
		}
		path_up( chk );
#if 0
fprintf( stdFp, "UP >> PATH %s\n", path );
#endif
	}
#if 1
	return check_entry( path, fd, chk );
#else
	return 0;
#endif
}

static inline int
repair_data( char * path, unsigned long * t1, struct stat * s1,
		int fd, check_t * chk,
			unsigned long t2, struct stat * s2,
			unsigned long of, unsigned long ln )
{
	int state = -1;

	switch( t2 )
	{
		case TYPE_REG:
			state = repair_reg( path, fd, chk,
						s2->st_mode, of, ln );
		break;

		case TYPE_LNK:
			state = repair_link( path, fd, chk, of, ln );

		break;

		case TYPE_DIR:
			state = repair_dir( path, fd, chk );
		break;

		default:
			goto out;
	}

	state = repair_attribute( path, t1, s1, t2, s2 );
out:
	return state;
}

int
create_dir( char * path, unsigned long * t1, struct stat * s1,
		int fd, check_t * chk,
			unsigned long t2, struct stat * s2,
			unsigned long of, unsigned long ln )
{
	int		state = -1;
	unsigned long	o = 0;
	unsigned long	t, nt;
	struct stat	s, ns;
	static char	n[256];

#if 1
	fprintf( stdFp, "Create directory : target %s ", path );
#endif
	if ( mkdir( path, s2->st_mode ) != -1 )
	{
#if 1
		fprintf( stdFp, "done.\n" );
#endif
	}
	else
	{
#if 1
		fprintf( stdFp, "error.\n" );
#endif
		goto out;
	}
	state = repair_attribute( path, t1, s1, t2, s2 );

	while ( next_entry( fd, chk, &o, &t, &s, n ) != -1 && strlen( n ) )
	{
#if 0
fprintf( stdFp, "<<<%s[%u]>%X>%X:%d:%d:%u>\n",
		n, o, t, s.st_mode, s.st_uid, s.st_gid, s.st_size );
#endif
		path_down( chk, n );
#if 0
fprintf( stdFp, "DOWN >> PATH %s\n", path );
#endif
		if ( strcmp( n, "." ) && strcmp( n, ".." ) )
		{
			nt = NO_EXISTS;
			memset( &ns, 0, sizeof( struct stat ) );
			switch( t )
			{
				case TYPE_REG:
				case TYPE_LNK:
					state = repair_data( path, &nt, &ns,
						fd, chk, t, &s, 0, (~0) );
				break;

				case TYPE_DIR:
					state = create_dir( path, &nt, &ns,
						fd, chk, t, &s, 0, (~0) );
				break;
			}
		}
		path_up( chk );
#if 0
fprintf( stdFp, "UP >> PATH %s\n", path );
#endif
	}

out:
	return	state;
}

static inline int
repair_create( char * path, unsigned long * t1, struct stat * s1,
		int fd, check_t * chk,
		unsigned long t2, struct stat * s2 )
{
	int		state = -1;

	switch ( t2 )
	{
		case TYPE_REG:
		case TYPE_LNK:
			state = repair_data( path, t1, s1,
						fd, chk, t2, s2, 0, (~0) );
		break;

		case TYPE_DIR:
			state = create_dir( path, t1, s1,
						fd, chk, t2, s2, 0, (~0) );
		break;
	}

	return state;
}

static inline int
connect_repair( unsigned long sip )
{
	struct sockaddr_in	sock;

	sock.sin_family = PF_INET;
	sock.sin_addr.s_addr = sip;
#if 0
	sock.sin_port = TARGET_REPAIR_PORT;
#else
	sock.sin_port = htons( repair_port );
#endif

	return tcp_connect( &sock );
}

static inline int
command_repair( int fd, check_t * chk )
{
	unsigned long	rt, ss, so, sl;
	int		rfd, state = -1;
	unsigned long	t, st;
	struct stat	s;
	pid_t		p;

	chk->recvbuff.offset = 0;
	chk->recvbuff.length = to_bytelen( 3 );
	if ( receive_buffer( fd, &chk->recvbuff ) < to_bytelen( 3 ) )
		goto out_err;
	rt = get_hlong( &chk->recvbuff );
	ss = get_nlong( &chk->recvbuff );
	sl = get_hlong( &chk->recvbuff );

	chk->workbuff.offset = 0;
	chk->workbuff.length = quad_u32len( sl );
	if ( receive_buffer( fd, &chk->workbuff ) < quad_u32len( sl ) )
		goto out_err;

	chk->workbuff.buffer[ sl++ ] = '/';
	strcpy( &( chk->workbuff.buffer[ sl ] ), chk->targ_pt );
	sl = strlen( chk->workbuff.buffer );

#if 1
fprintf( stdFp, "RECOVERY REQUEST : TYPE %d\nPATH %s\n",
		rt, chk->workbuff.buffer );
fflush( stdFp );
#endif

	if ( ( p = fork() ) > 0 )
	{
		return send_status( fd,  STATUS_OK );
	}
	else if ( p < 0 )
	{
		fprintf( stdFp, "Child fork exception.\n" );
		goto out_err;
	}

	if ( ( rfd = connect_repair( ss ) ) == -1 )
	{
		fprintf( errFp, "Repair connect exception.\n" );
		goto out;
	}

	if ( get_stat( rfd, chk, chk->workbuff.buffer, &t, &s ) == -1 )
		goto out_disconnect;

	if ( ( rt == REC_REMOVE && t == NO_EXISTS ) ||
		( rt & REC_REMOVE && t != chk->targ_type ) )
		state = repair_remove( chk->serv_ex,
					&chk->targ_type, &chk->targ_st );
	if ( rt & REC_CREATE )
		state = repair_create( chk->serv_ex,
					&chk->targ_type, &chk->targ_st,
					rfd, chk, t, &s );
	if ( rt & REC_ATTR )
		state = repair_attribute( chk->serv_ex,
						&chk->targ_type,
						&chk->targ_st, t, &s );
	if ( rt & REC_DATA )
		state = repair_data( chk->serv_ex,
					&chk->targ_type, &chk->targ_st,
					rfd, chk, t, &s,
					chk->targ_of, chk->targ_ln );

out_disconnect:
	if ( send_status( rfd, COMMAND_EXIT ) != -1 )
		receive_status( rfd, &st );
	close( rfd );
out:
	fflush( stdFp );

	exit( 0 );

out_err:
	return state;
}

static inline int
check_loop( int fd )
{
	check_t chk;

	chk.pathbuff.buffer = chk.pbuff;
	chk.workbuff.buffer = chk.wbuff;
	chk.sendbuff.buffer = chk.sbuff;
	chk.recvbuff.buffer = chk.rbuff;

	chk.pathbuff.offset = chk.pathbuff.length = 0;
	while( 1 )
	{
		unsigned long	com;
		int		status = -1;

		chk.recvbuff.offset = 0;
		chk.recvbuff.length = U32_LENGTH;

		if ( receive_buffer( fd, &chk.recvbuff ) < U32_LENGTH )
			break;

		switch( ( com = get_hlong( &chk.recvbuff ) ) )
		{
			case COMMAND_AUTH:
#if 0
			fprintf( stdFp, "CHECK_COMMAND_AUTH : " );
#endif
				status = command_auth( fd, &chk );
			break;

			case COMMAND_STAT:
#if 0
			fprintf( stdFp, "CHECK_COMMAND_STAT : " );
#endif
				status = command_stat( fd, &chk );
			break;

			case COMMAND_REPAIR:
#if 0
			fprintf( stdFp, "CHECK_COMMAND_REPAIR : " );
#endif
				status = command_repair( fd, &chk );
			break;

			case COMMAND_EXIT:
#if 0
			fprintf( stdFp, "CHECK_COMMAND_EXIT : BYE!\n" );
#endif
				return	send_status( fd, STATUS_OK );
			default:
			fprintf( errFp,
				"CHECK_COMMAND_FAILED( %d ) : BYE!\n", com );
				return	send_status( fd, STATUS_NG );
		}
		if ( status < 0 ) break;
	}
	return send_status( fd, STATUS_NG );
}

void
check_main( int sock_fd )
{
	int			fd = 0;
	struct sockaddr_in	sock;

	signal(SIGCHLD, SIG_IGN);
	do
	{
		fflush( stdFp );
		if ( ( fd = tcp_accept( sock_fd, &sock ) ) != -1 )
		{
			time_t	t = time( NULL );
			fprintf( stdFp, "CHECK CONNECTION : %s : %s",
					inet_ntoa( sock.sin_addr.s_addr ),
					ctime( &t ) );
			check_loop( fd );
			close( fd );
		}
	}
	while( !sig_num );
}
