/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#ifndef MSGMAX
#define	MSGMAX		4080
#endif

#define	POOL_MAX	16
#define	MSG_SIZE		( ( MSGMAX + 0xfff ) & 0x1000 )

static int	msg_qid;
static int	msg_size[ POOL_MAX ];
static char	msg_pool[ POOL_MAX ][ MSG_SIZE ];

static inline void *
MSG_BUFF( i )
{
	return	( ( void * )msg_pool[ i ] );
}

static inline int
U32L( int i )
{
	return	( i << 2 );
}

static inline char *
MSG_OFFS( i )
{
	return	( &msg_pool[ i ][ U32L( 2 ) ] );
}

static const int flush_now = 120;
static int flush_counter = 0;

static inline void
flush_pool()
{
	int	i;

	for( i = 0; i < POOL_MAX; i++ )
		if ( !msg_size[ i ] )	msg_size[ i ] = -1;
	flush_counter = 0;
}

void
init_pool( int qid )
{
	int	i = 0;

	msg_qid = qid;
	while( i < POOL_MAX )
		msg_size[ i++ ] = -1;
}

void
release_pool( char * buf )
{
	int	i = 0;

	while( i < POOL_MAX && msg_pool[ i ] != buf ) i++;

	if ( i == POOL_MAX )	return;

	msg_size[ i ] = 0;
}

int
scan_pool( char ** buf, int *len )
{
	int	i = 0;

	while( i < POOL_MAX && msg_size[ i ] <= 0 ) i++;

	if ( i == POOL_MAX )	return 0;

	*len = msg_size[ i ];
	*buf = msg_pool[ i ];

	return *len;
}

static inline int
msg_check_str( char * buf, int len )
{
	int slen = -1, qlen;

	if ( len < U32L( 1 ) )
		return	-1;

	slen = ntohl( *( ( unsigned long * )buf ) );
	qlen = ( ( slen + 3 ) & (~3) );

	len -= U32L( 1 );
	buf += U32L( 1 );
	if ( qlen > len )	return	-1;

	while( slen < qlen )	buf[ slen++ ] = '\0';
	len -= qlen;
out:
	return len;
}

static inline int
msg_check_serv_one( char * buf, int len )
{
	if ( len < U32L( 2 ) )	goto	out_err;
	buf += U32L( 1 );
	len -= U32L( 1 );
	return	msg_check_str( buf, len );
out_err:
	return -1;
}

static inline int
msg_check_serv( char * buf, int len )
{
	int	i = 0, n = ntohl( *( ( unsigned long * )buf ) );
	int	clen;

	buf += U32L( 1 );
	len -= U32L( 1 );
	if ( len < U32L( n ) )	goto	out_err;

	clen = len;
	while( i < n )
	{
		clen = msg_check_serv_one( &buf[ len - clen ], clen );
		if ( clen <= 0 ) goto	out_err;
		i++;
	}

	return clen;
out_err:
	return -1;
}

static inline int
msg_check_targ_one( char * buf, int len )
{
	if ( len < U32L( 2 ) )	goto	out_err;
	buf += U32L( 1 );
	len -= U32L( 1 );
	len = msg_check_str( buf, len );
	if ( len == -1 || len < U32L( 2 ) )
		goto	out_err;
	len -= U32L( 2 );
	return len;
out_err:
	return -1;
}

static inline int
msg_check_targ( char * buf, int len )
{
	int	i = 0, n = ntohl( *( ( unsigned long * )buf ) );
	int	clen;

	buf += U32L( 1 );
	len -= U32L( 1 );
	if ( len < U32L( n ) )	goto	out_err;

	clen = len;
	while( i < n )
	{
		clen = msg_check_targ_one( &buf[ len - clen ], clen );
		if ( clen < 0 ) goto	out_err;
		i++;
	}

	return clen;
out_err:
	return -1;
}

static inline int
msg_check( char * buf, int len )
{
	int		wlen = len;
	
	if ( wlen < U32L( 6 ) )	goto	out_err;
	wlen -= U32L( 5 );
	if ( ( wlen = msg_check_serv( &buf[ len - wlen ], wlen ) ) == -1 )
		goto	out_err;
	if ( wlen < U32L( 1 ) ) goto	out_err;
	if ( msg_check_targ( &buf[ len - wlen ], wlen ) )
		goto	out_err;
	return len;
out_err:
	return -1;
}

static inline int
pool_msg_one( int i )
{
	int	j;

try_next:
	msg_size[ i ] =
		msgrcv( msg_qid, MSG_BUFF( i ), MSGMAX, 0, IPC_NOWAIT );

	if (  msg_size[ i ] > 0 )	flush_counter = 0;
	else if ( ++flush_counter == flush_now )
		flush_pool();
	else	goto	out;

	msg_size[ i ] = msg_check( MSG_BUFF( i ), msg_size[ i ] );

	if ( msg_size[ i ] == -1 )	goto out;

	j = 0;
	while( j < POOL_MAX )
	{
		if ( i != j &&  msg_size[ j ] >= 0 &&
		     !memcmp( MSG_OFFS( i ), MSG_OFFS( j ),
				msg_size[ i ] - U32L( 2 ) ) )
			goto	try_next;
		j++;
	}
out:
	return msg_size[ i ];
}

static inline int
pool_msg()
{
	int	i = 0, c = 0;

	while( i < POOL_MAX )
	{
		if ( msg_size[ i ] < 0 )
		{
			if ( pool_msg_one( i ) < 0 )
				goto	out;
			else if (  msg_size[ i ] > 0 )
				c++;
		}
		i++;
	}
	i = 0;
	while( i < POOL_MAX )
	{
		if ( !msg_size[ i ] )
		{
			if ( pool_msg_one( i ) < 0 )
				goto	out;
			else if (  msg_size[ i ] > 0 )
				c++;
		}
		i++;
	}
out:
	return c;
}

int
next_pool( char ** buf, int *len )
{
	int	l, p;

	if ( !buf || !len )
		return -1;

	l = scan_pool( buf, len );

	p = pool_msg();

	if ( l == 0 && p > 0 )	l = scan_pool( buf, len );

	return l;
}
