/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#ifndef __RECOVERY_INCLUDE__
#define __RECOVERY_INCLUDE__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "buffer.h"

#if 0
#define	IPC_DEFAULT_KEY		(key_t)2048
#define TARGET_CHECK_PORT	htons(2048)
#define TARGET_REPAIR_PORT	htons(2049)
#define CLIENT_AUTH_PORT	htons(111)
#else
key_t	recovery_key;
short	check_port;
short	repair_port;
short	pmap_port;
#endif

#define TCP_TIMEOUT_ONE		10000
#define TCP_TIMEOUT_TRY		500

extern	FILE *stdFp;
extern	FILE *errFp;

extern	volatile sig_atomic_t	sig_num;

/*
 * tcpsocket.c
 * functions
 */
int	tcp_canread( int fd );
int	tcp_canwrite( int fd );

int	tcp_bind( struct sockaddr_in * sock );
int	tcp_accept( int fd, struct sockaddr_in * sock );
int	tcp_connect( struct sockaddr_in * sock );

int	tcp_read_one( int fd, unsigned char * buf, int len );
int	tcp_read( int fd, unsigned char * buf, int len );
int	tcp_write_one( int fd, unsigned char * buf, int len );
int	tcp_write( int fd, unsigned char * buf, int len );

int	receive_buffer( int fd, buffer_t * buff );
int	send_buffer( int fd, buffer_t * buff );
int	send_status( int fd, unsigned long status );
int	receive_status( int fd, unsigned long * status );

/*
 * recovery.c
 * functions
 */
#define	STATUS_OK	0
#define	STATUS_NG	(~0)

void	recovery_main( int qid );

/*
 * check.c
 * functions
 */
#define	REC_CREATE	1
#define	REC_REMOVE	2
#define	REC_ATTR	4
#define	REC_DATA	8

#define	COMMAND_AUTH	0
#define	COMMAND_STAT	1
#define	COMMAND_ENTRY	2
#define	COMMAND_EXISTS	3
#define	COMMAND_DATA	4
#define COMMAND_REPAIR	10
#define	COMMAND_EXIT	(~0)

void	check_main( int sock_fd );

/*
 * repair.c
 * functions
 */
void	repair_main( int sock_fd );

/*
 * msgpool.c
 * functions
 */
void	init_pool( int qid );
int	next_pool( char ** buf, int * len );

/*
 * majority.c
 * functions
 */
int	majority( recovery_t * rec );

/*
 * checksum.c
 * functions
 */
unsigned long checksum( unsigned long type, char * file,
			unsigned long off, unsigned long len );

/*
 * auth.c
 * functions
 */
int	auth_client( char *export, unsigned long int sip, unsigned long int cip );

/*
 * stat type functions
 */
#define	NO_EXISTS	0
#define	TYPE_REG	1
#define	TYPE_DIR	2
#define	TYPE_LNK	3
#define	TYPE_OTH	4

static inline unsigned long
filetype( unsigned long mode )
{
	if ( !mode )
		return  NO_EXISTS;
	else if ( S_ISREG( mode ) )
		return  TYPE_REG;
	else if ( S_ISDIR( mode ) )
		return  TYPE_DIR;
	else if ( S_ISLNK( mode ) )
		return  TYPE_LNK;

	return  TYPE_OTH;
}

static inline void
readstat( char *p, unsigned long * t, struct stat  * s )
{
	if ( lstat( p, s ) == -1 )
	{
		*t = NO_EXISTS;
		memset( s, 0, sizeof( struct stat ) );
	}
	else
	{
		if ( ( *t = filetype( s->st_mode ) ) == TYPE_DIR )
			s->st_size = s->st_nlink;
	}
}

#endif
