/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

static inline int
command_stat( int fd, repair_t * rep )
{
	int             slen;

	rep->targ_ct = 0;
	
        rep->recvbuff.offset = 0;
        rep->recvbuff.length = to_bytelen( 1 );
        if ( receive_buffer( fd, &rep->recvbuff ) < to_bytelen( 1 ) )
        	goto out_err;

        slen = get_hlong( &rep->recvbuff );
        rep->pathbuff.offset = 0;
        rep->pathbuff.length = quad_u32len( slen );
        if ( receive_buffer( fd, &rep->pathbuff ) < quad_u32len( slen ) )
        	goto out_err;
	rep->pathbuff.buffer[ slen ] = '\0';

	readstat( rep->pathbuff.buffer, &rep->targ_type, &rep->targ_st );

#if 1
fprintf( stdFp, "TARGET : %s\n", rep->pathbuff.buffer );
#endif

	rep->sendbuff.offset = rep->sendbuff.length = 0;
        if ( !free_u32length( &rep->sendbuff, 6 ) )
                goto out_err;
        set_hlong( &rep->sendbuff, rep->targ_type );
        set_hlong( &rep->sendbuff, rep->targ_st.st_mode );
        set_hlong( &rep->sendbuff, rep->targ_st.st_uid );
        set_hlong( &rep->sendbuff, rep->targ_st.st_gid );
        set_hlong( &rep->sendbuff, rep->targ_st.st_size );

        if ( send_buffer( fd, &rep->sendbuff ) < to_bytelen( 5 ))
                goto out_err;
	rep->targ_pt = rep->pathbuff.buffer;

	if ( rep->targ_type != NO_EXISTS )
		rep->targ_pt = rep->pathbuff.buffer;
	else
		rep->targ_pt = NULL;

	return 0;
out_err:
        return -1;
}

static inline int
command_data_reg( int fd, repair_t * rep, unsigned long of, unsigned long ln )
{
	int		rfd, len, state = -1;

#if 0
fprintf( stdFp, "TARGET %s FILE [ %u - %u ]\n", rep->targ_pt, of, ln );
#endif

	if ( ( rfd = open( rep->targ_pt, O_RDONLY ) ) == -1 )
		goto out;
	if ( lseek( rfd, of, SEEK_SET ) == -1 )
		goto out_close;

	rep->sendbuff.offset = rep->sendbuff.length = 0;
	set_hlong( &rep->sendbuff, STATUS_OK );
	set_hlong( &rep->sendbuff, ln );
	len = rep->sendbuff.length;
	if ( send_buffer( fd, &rep->sendbuff ) < len )
		goto out_close;

	len = ( ( ln < BUFFER_SIZE_BYTE ) ?  ln : BUFFER_SIZE_BYTE );

	while( len && ( len = read( rfd, rep->sendbuff.buffer, len ) ) > 0 )
	{
		if ( tcp_write( fd, rep->sendbuff.buffer, len ) < len )
			goto out_close;
		if ( ( ln -= len ) <= 0 )
			break;
		len = ( ( len < ln ) ? len : ln );
	}
	state = 0;

out_close:
	close( rfd );
out:
	return	state;
}

static inline int
command_data_link( int fd, repair_t * rep )
{
	int		len;
	unsigned long *	buff = ( unsigned long * )rep->sendbuff.buffer;

	if ( ( len = readlink( rep->targ_pt, ( char * )&buff[ 2 ],
			BUFFER_SIZE_BYTE - to_bytelen( 2 ) ) ) == -1 )
		goto out_err;
	buff[ 0 ] = htonl( STATUS_OK );
	buff[ 1 ] = htonl( len );

#if 0
fprintf( stdFp, "TARGET %s LINK [ 0 - %u ]\n", rep->targ_pt, len );
#endif

	return tcp_write( fd, rep->sendbuff.buffer, to_bytelen( 2 ) + len );

out_err:
        return -1;
}

static inline int
command_data( int fd, repair_t * rep )
{
	unsigned long	of, ln, l;

	if ( !rep->targ_pt )
		goto out_err;

        rep->recvbuff.offset = 0;
        rep->recvbuff.length = to_bytelen( 2 );
        if ( receive_buffer( fd, &rep->recvbuff ) < to_bytelen( 2 ) )
        	goto out_err;
        of = get_hlong( &rep->recvbuff );
        ln = get_hlong( &rep->recvbuff );

	l = rep->targ_st.st_size - of;
	ln = ( ( ln < l ) ? ln : l );

	if ( rep->targ_type == TYPE_REG )
		return command_data_reg( fd, rep, of, ln );
	if ( rep->targ_type == TYPE_LNK )
		return command_data_link( fd, rep );
out_err:
        return -1;
}

static inline int
path_down( repair_t * rep, char * path )
{
	int	tn = strlen( rep->targ_pt ), pn = strlen( path );

	if ( ( tn + pn + 2 ) > BUFFER_SIZE_BYTE )
		return -1;
	rep->targ_pt[ tn++ ] = '/';
	strcpy( &( rep->targ_pt[ tn ] ), path );
	rep->targ_ct ++;

	return 0;
}

static inline int
path_up( repair_t * rep )
{
	int	tn = strlen( rep->targ_pt );

	if ( ( --(rep->targ_ct) ) < 0 )	return -1;
	while ( tn > 0 && rep->targ_pt[ --tn ] != '/' );
	if ( tn < 0 )		return -1;
	rep->targ_pt[ tn ] = '\0';

	return 0;
}

static inline int
command_exists(  int fd, repair_t * rep )
{
	struct stat	st;
	unsigned long	ty, slen, qlen;

	if ( !rep->targ_pt )
		goto out_err;
	if ( receive_status( fd, &slen ) == -1 )
		goto out_err;
	qlen = quad_u32len( slen );
        rep->recvbuff.offset = 0;
        rep->recvbuff.length = qlen;
        if ( receive_buffer( fd, &rep->recvbuff ) < qlen )
        	goto out_err;
	rep->recvbuff.buffer[ slen ] = '\0';

	path_down( rep, rep->recvbuff.buffer );
#if 0
fprintf( stdFp, "TARGET %s\n", rep->targ_pt );
#endif
	readstat( rep->targ_pt, &ty, &st );
	path_up( rep );

	rep->sendbuff.offset = rep->sendbuff.length = 0;
        set_hlong( &rep->sendbuff, ty );
        set_hlong( &rep->sendbuff, st.st_mode );
        set_hlong( &rep->sendbuff, st.st_uid );
        set_hlong( &rep->sendbuff, st.st_gid );
        set_hlong( &rep->sendbuff, st.st_size );
	qlen = rep->sendbuff.length;

        if ( send_buffer( fd, &rep->sendbuff ) < qlen )
                goto out_err;

	return  0;
out_err:
        return -1;
}

static inline int
command_entry( int fd, repair_t * rep )
{
	DIR *		d;
	struct dirent *	e;
	struct stat	s;
	unsigned long	o, t, l;
	char *		n;

	if ( !rep->targ_pt )
		goto out_err;

#if 0
fprintf( stdFp, "TARGET %s\n", rep->targ_pt );
#endif

	if ( receive_status( fd, &o ) == -1 )
		goto out_err;

	if ( o ) path_up( rep );

	if ( !( d = opendir( rep->targ_pt ) ) )
		goto out_err;

	if ( o ) seekdir( d, o );

	if ( !( e = readdir( d ) ) )
	{
		n = "";
		t = NO_EXISTS;
		memset( &s, 0, sizeof( struct stat ) );

		readstat( rep->targ_pt, &rep->targ_type, &rep->targ_st );
	}
	else
	{
		n = e->d_name;

		path_down( rep, n );

		readstat( rep->targ_pt, &rep->targ_type, &rep->targ_st );
		memcpy( &s, &rep->targ_st, sizeof( struct stat ) );
		t = rep->targ_type;

		o = telldir( d );
	}

	rep->sendbuff.offset = rep->sendbuff.length = 0;
        set_hlong( &rep->sendbuff, t );
        set_hlong( &rep->sendbuff, s.st_mode );
        set_hlong( &rep->sendbuff, s.st_uid );
        set_hlong( &rep->sendbuff, s.st_gid );
        set_hlong( &rep->sendbuff, s.st_size );
        set_hlong( &rep->sendbuff, o );
        set_hstr( &rep->sendbuff, n );
	l = rep->sendbuff.length;
	closedir( d );

        if ( send_buffer( fd, &rep->sendbuff ) < l )
                goto out_err;

        return 0;
out_err:
        return -1;
}

static inline int
repair_loop( int fd )
{
	repair_t	rep;

	rep.pathbuff.buffer = rep.pbuff;
	rep.sendbuff.buffer = rep.sbuff;
	rep.recvbuff.buffer = rep.rbuff;

        while( 1 )
        {
		unsigned long	com;
                int     	status = -1;

                rep.recvbuff.offset = 0;
                rep.recvbuff.length = U32_LENGTH;

                if ( receive_buffer( fd, &rep.recvbuff ) < U32_LENGTH )
			break;

                switch( ( com = get_hlong( &rep.recvbuff ) ) )
                {
                        case COMMAND_STAT:
#if 0
                        fprintf( stdFp, "REPAIR_COMMAND_STAT : " );
#endif
                                status = command_stat( fd, &rep );
                        break;

                        case COMMAND_DATA:
#if 0
                        fprintf( stdFp, "REPAIR_COMMAND_DATA : " );
#endif
                                status = command_data( fd, &rep );
                        break;

                        case COMMAND_EXISTS:
                                status = command_exists( fd, &rep );
                        break;

                        case COMMAND_ENTRY:
#if 0
                        fprintf( stdFp, "REPAIR_COMMAND_ENTRY : " );
#endif
                                status = command_entry( fd, &rep );
                        break;

                        case COMMAND_EXIT:
#if 0
                        fprintf( stdFp, "REPAIR_COMMAND_EXIT : BYE!\n" );
#endif
                                return  send_status( fd, STATUS_OK );

			default:
        		fprintf( errFp,
				"REPAIR_COMMAND_FAILED( %d ): BYE!\n", com );
                                return  send_status( fd, STATUS_NG );
		}
                if ( status < 0 ) break;
	}
        return send_status( fd, STATUS_NG );
}


void
repair_main( int sock_fd )
{
	pid_t			pid;
        int                     fd = 0;
        struct sockaddr_in      sock;

	signal(SIGCHLD, SIG_IGN);
        do
        {
		fflush( stdFp );
                if ( ( fd = tcp_accept( sock_fd, &sock ) ) != -1 )
		{
			time_t	t = time( NULL );
			fprintf( stdFp, "REPAIR CONNECTION : %s : %s",
					inet_ntoa( sock.sin_addr.s_addr ),
					ctime( &t ) );
			if ( !( pid = fork() ) )
			{
                        	repair_loop( fd );
                        	close( fd );
				exit( 0 );
			}
			close( fd );
                }
        }
        while( !sig_num );
}
